"""
Transpile the regular expressions to instructions of a Virtual Machine.

The implementation in the standard library has exponential time complexity, so it was
a major blocker for most of the practical inputs. For example, see this bug report:
https://gcc.gnu.org/bugzilla/show_bug.cgi?id=93502

The virtual machine is based on Ken Thompson's approach published in:
Thompson, K., "Regular expression search algorithm", ACM 11(6) (June 1968)

We followed a very clear and concise blog post which described it in detail:
https://swtch.com/~rsc/regexp/regexp2.html

The ideas for additional instructions were taken from:
https://www.codeproject.com/Articles/5256833/Regex-as-a-Tiny-Threaded-Virtual-Machine
"""
import dataclasses
import io
from typing import (
    Final,
    cast,
    List,
    Optional,
    Union,
    Tuple,
    TextIO,
    MutableMapping,
    Sequence,
)

from icontract import require

from aas_core_codegen.common import assert_never, pairwise
from aas_core_codegen.parse import retree as parse_retree, tree as parse_tree


class Character(str):
    """Represent a single character."""

    @require(lambda text: len(text) == 1)
    def __new__(cls, text: str) -> "Character":
        return cast(Character, text)


@dataclasses.dataclass
class InstructionChar:
    """Match a single character."""

    character: Character


class Range:
    """Define a character range."""

    first: Final[Character]
    last: Final[Character]

    # fmt: off
    @require(
        lambda first, last: ord(first) <= ord(last),
        "Range boundaries must be sorted."
    )
    # fmt: on
    def __init__(self, first: Character, last: Character) -> None:
        """Initialize with the given values."""
        self.first = first
        self.last = last

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.first!r}, {self.last!r})"


def check_ranges_sorted_and_non_overlapping(ranges: Sequence[Range]) -> Optional[str]:
    """
    Check that the ranges' boundaries are sorted and non-overlapping.

    If there are no errors, return ``None``. Otherwise, return a message explaining
    what precisely was not satisfied.
    """
    for this_range, next_range in pairwise(ranges):
        if ord(this_range.first) >= ord(next_range.last):
            return (
                f"The range {this_range} and its next range {next_range} "
                f"are not in sorted order."
            )

        if ord(this_range.last) >= ord(next_range.first):
            return f"The range {this_range} and its next range {next_range} overlap."

    return None


class InstructionSet:
    """Match a set of characters."""

    ranges: Final[Sequence[Range]]

    @require(lambda ranges: check_ranges_sorted_and_non_overlapping(ranges) is None)
    def __init__(self, ranges: Sequence[Range]) -> None:
        """Initialize with the given values."""
        self.ranges = ranges

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.ranges!r})"


class InstructionNotSet:
    """Match an out-of-set character."""

    ranges: Final[Sequence[Range]]

    @require(lambda ranges: check_ranges_sorted_and_non_overlapping(ranges) is None)
    def __init__(self, ranges: Sequence[Range]) -> None:
        """Initialize with the given values."""
        self.ranges = ranges

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.ranges!r})"


@dataclasses.dataclass
class InstructionAny:
    """Match any character."""


@dataclasses.dataclass
class InstructionMatch:
    """Stop the thread and signal that we found a match."""


@dataclasses.dataclass
class InstructionJump:
    """Jump to the indicated position in the program."""

    target: int


@dataclasses.dataclass
class InstructionSplit:
    """Split the program in two threads, both jumping to different locations."""

    first_target: int
    second_target: int


@dataclasses.dataclass
class InstructionEnd:
    """Match the end-of-input."""


@dataclasses.dataclass
class _InstructionNoop:
    """
    Represent a no-operation instruction which does nothing.

    This is used only as a place-holder during the translation.
    """


Instruction = Union[
    InstructionChar,
    InstructionSet,
    InstructionNotSet,
    InstructionAny,
    InstructionMatch,
    InstructionJump,
    InstructionSplit,
    InstructionEnd,
]


# NOTE (mristin):
# The classes ``_Leaf`` and ``_Node`` correspond to the translation phase, while
# the classes ``Leaf`` and ``Node`` are used to represent the final result.


@dataclasses.dataclass
class _Leaf:
    """Represent a leaf node in the nested instruction tree during compilation."""

    instruction: Union[Instruction, _InstructionNoop]
    label: Optional[int] = None


@dataclasses.dataclass
class _Node:
    """Represent nested instructions during compilation."""

    re_node: parse_retree.Node
    children: List["_NodeOrLeaf"]


_NodeOrLeaf = Union[_Node, _Leaf]


@dataclasses.dataclass
class Leaf:
    """Represent a leaf node in the nested instruction tree."""

    instruction: Instruction
    label: Optional[int] = None


@dataclasses.dataclass
class Node:
    """Represent nested instructions related to a part of the regular expression."""

    re_node: parse_retree.Node
    children: List["NodeOrLeaf"]


NodeOrLeaf = Union[Node, Leaf]


class _CheckForFormattedValue(parse_retree.PassThroughVisitor):
    """Check that no ``FormattedValue`` is contained in a pattern."""

    def __init__(self) -> None:
        self.has_formatted_value = False

    def visit_term(self, node: parse_retree.Term) -> None:
        if isinstance(node.value, parse_tree.FormattedValue):
            self.has_formatted_value = True

        super().visit_term(node)


class _CheckForNonGreedyQuantifiers(parse_retree.PassThroughVisitor):
    """Check for presence of non-greedy quantifiers in a pattern."""

    def __init__(self) -> None:
        self.has_non_greedy_quantifiers = False

    def visit_quantifier(self, node: parse_retree.Quantifier) -> None:
        if node.non_greedy:
            self.has_non_greedy_quantifiers = True

        super().visit_quantifier(node)


class _RegexRenderer(parse_retree.Renderer):
    """
    Render regex patterns for readable comments.

    In contrast to :py:class:`parse_retree.Renderer`, we also render
    the :py:class:`parse_retree.Char` as we need to cover the granularity of a single
    character in this module.
    """

    def transform_char(
        self, node: parse_retree.Char
    ) -> List[Union[str, parse_tree.FormattedValue]]:
        return self.char_to_str_and_escape_or_encode_if_necessary(
            node=node, escaping=parse_retree.Renderer._ESCAPING_IN_CHARACTER_LITERALS
        )


_REGEX_RENDERER = _RegexRenderer()


def _render_re_node(re_node: parse_retree.Node) -> str:
    """Render the pattern as a string for comments."""
    parts = _REGEX_RENDERER.transform(re_node)
    assert all(
        isinstance(part, str) for part in parts
    ), f"Expected all rendered parts to be strings, but got: {parts}"

    return "".join(parts)  # type: ignore


class _Translator(parse_retree.Transformer[_Node]):
    """Translate the regular expression to nested instructions."""

    def __init__(self) -> None:
        self._next_label = 0

    def _obtain_label(self) -> int:
        """
        Return the current next label, and increment it for the next call.

        Do not fiddle with :py:attr:`_next_label` yourself; use this function.
        """
        result = self._next_label
        self._next_label += 1
        return result

    def transform_union_expr(self, node: parse_retree.UnionExpr) -> _Node:
        if len(node.uniates) == 0:
            return _Node(re_node=node, children=[_Leaf(instruction=_InstructionNoop())])
        elif len(node.uniates) == 1:
            return self.transform(node.uniates[0])
        else:
            pass

        children = []  # type: List[_NodeOrLeaf]

        final_label = self._obtain_label()

        for i, uniate in enumerate(node.uniates):
            if i < len(node.uniates) - 1:
                l0 = self._obtain_label()
                l1 = self._obtain_label()

                children.append(
                    _Leaf(InstructionSplit(first_target=l0, second_target=l1))
                )

                children.append(_Leaf(_InstructionNoop(), label=l0))

                children.append(self.transform(uniate))

                children.append(_Leaf(InstructionJump(target=final_label)))

                children.append(_Leaf(_InstructionNoop(), label=l1))
            else:
                children.append(self.transform(uniate))

                children.append(_Leaf(_InstructionNoop(), label=final_label))

        return _Node(re_node=node, children=children)

    def transform_concatenation(self, node: parse_retree.Concatenation) -> _Node:
        if len(node.concatenants) == 0:
            return _Node(re_node=node, children=[_Leaf(instruction=_InstructionNoop())])

        elif len(node.concatenants) == 1:
            return self.transform(node=node.concatenants[0])

        else:
            return _Node(
                re_node=node,
                children=[
                    self.transform(concatenant) for concatenant in node.concatenants
                ],
            )

    def transform_symbol(self, node: parse_retree.Symbol) -> _Node:
        if node.kind is parse_retree.SymbolKind.START:
            raise AssertionError(
                "We expect that the caller skipped the start anchor "
                "as we always expect the patterns to be anchored at start, and there "
                "is no matching to be done against the start anchor. We decided "
                "against introduction of a no-op instruction since that only eats up "
                "resources, which we can avoid with a bit of smartness in "
                "the transpilation phase."
            )

        elif node.kind is parse_retree.SymbolKind.END:
            return _Node(re_node=node, children=[_Leaf(instruction=InstructionEnd())])

        elif node.kind is parse_retree.SymbolKind.DOT:
            return _Node(
                re_node=node,
                children=[_Leaf(instruction=InstructionAny())],
            )

        else:
            assert_never(node.kind)

    def transform_term(self, node: parse_retree.Term) -> _Node:
        assert not isinstance(node.value, parse_tree.FormattedValue), (
            "Unexpected formatted value in the regular expression to be "
            "transformed into a program for Regex Virtual Machine. "
            "This should have been checked before. "
            f"The formatted value is: {parse_tree.dump(node.value)}"
        )

        if node.quantifier is not None:
            if node.quantifier.non_greedy:
                raise AssertionError(
                    "(mristin, 2024-06-04) Only non-greedy quantifiers are currently "
                    "translated to a program for a RegEx virtual machine. We did not "
                    "cover non-greedy quantifiers for simplicity, as we currently have "
                    "no meta-model where they are required. The presence of non-greedy "
                    "quantifiers should have been caught before, as we explicitly "
                    "check for them when transpiling a meta-model. Please report "
                    "this exception to the developers as a bug."
                )

            if node.quantifier.minimum == 1 and node.quantifier.maximum == 1:
                return self.transform(node.value)

            children = []  # type: List[_NodeOrLeaf]

            if node.quantifier.maximum is not None:
                for _ in range(0, node.quantifier.minimum):
                    children.append(self.transform(node.value))

                optional_count = node.quantifier.maximum - node.quantifier.minimum
                if optional_count > 0:
                    final_label = self._obtain_label()

                    for _ in range(optional_count):
                        l1 = self._obtain_label()
                        children.append(
                            _Leaf(
                                InstructionSplit(
                                    first_target=l1, second_target=final_label
                                )
                            )
                        )
                        children.append(_Leaf(_InstructionNoop(), label=l1))
                        children.append(self.transform(node.value))

                    children.append(_Leaf(_InstructionNoop(), label=final_label))
            else:
                if node.quantifier.minimum == 0:
                    l1 = self._obtain_label()
                    l2 = self._obtain_label()
                    final_label = self._obtain_label()

                    children.append(
                        _Leaf(
                            InstructionSplit(
                                first_target=l2, second_target=final_label
                            ),
                            label=l1,
                        )
                    )

                    children.append(_Leaf(_InstructionNoop(), label=l2))
                    children.append(self.transform(node.value))
                    children.append(_Leaf(InstructionJump(target=l1)))

                    children.append(_Leaf(_InstructionNoop(), label=final_label))

                else:
                    # NOTE (mristin):
                    # The last mandatory repetition will be used for the unbounded loop.
                    for _ in range(0, node.quantifier.minimum - 1):
                        children.append(self.transform(node.value))

                    l1 = self._obtain_label()
                    final_label = self._obtain_label()

                    children.append(_Leaf(_InstructionNoop(), label=l1))
                    children.append(self.transform(node.value))
                    children.append(
                        _Leaf(
                            InstructionSplit(first_target=l1, second_target=final_label)
                        )
                    )
                    children.append(_Leaf(_InstructionNoop(), label=final_label))

            return _Node(re_node=node, children=children)

        else:
            return self.transform(node.value)

    def transform_group(self, node: parse_retree.Group) -> _Node:
        return self.transform(node.union)

    def transform_char(self, node: parse_retree.Char) -> _Node:
        return _Node(
            re_node=node,
            children=[
                _Leaf(instruction=InstructionChar(character=Character(node.character)))
            ],
        )

    def transform_quantifier(self, node: parse_retree.Quantifier) -> _Node:
        raise AssertionError(
            f"Expected the quantifier to be already handled "
            f"in {_Translator.transform_term.__name__}. We should have never gotten "
            f"here."
        )

    def transform_char_set(self, node: parse_retree.CharSet) -> _Node:
        ranges = [
            Range(
                first=Character(re_range.start.character),
                last=(
                    Character(re_range.end.character)
                    if re_range.end is not None
                    else Character(re_range.start.character)
                ),
            )
            for re_range in node.ranges
        ]

        ranges.sort(key=lambda rng: rng.first)

        if node.complementing:
            return _Node(
                re_node=node,
                children=[_Leaf(instruction=InstructionNotSet(ranges=ranges))],
            )
        else:
            return _Node(
                re_node=node,
                children=[_Leaf(instruction=InstructionSet(ranges=ranges))],
            )

    def transform_range(self, node: parse_retree.Range) -> _Node:
        raise AssertionError(
            "Expected to handle a regex range within the character set"
        )

    def transform_regex(self, node: parse_retree.Regex) -> _Node:
        non_anchored_exception_message = (
            "(mristin, 2024-05-31): We expect all the patterns which need "
            "to be transpiled to instructions of the RegEx virtual machine "
            "to be anchored at the start (``^``) and at the end (``$``). Please "
            "consider re-writing your pattern with putting a prefix ``^.*`` if you "
            "want to match an arbitrary prefix, and a suffix ``.*$`` if you want to "
            "match an arbitrary suffix. "
            "If you really need this feature, please contact the developers. "
            f"The regular expression was: {_render_re_node(node)}"
        )

        if len(node.union.uniates) == 0 or len(node.union.uniates[0].concatenants) == 0:
            raise NotImplementedError(non_anchored_exception_message)

        first_term = node.union.uniates[0].concatenants[0]
        last_term = node.union.uniates[0].concatenants[-1]

        first_symbol_is_start = isinstance(first_term.value, parse_retree.Symbol) and (
            first_term.value.kind is parse_retree.SymbolKind.START
        )
        last_symbol_is_end = isinstance(last_term.value, parse_retree.Symbol) and (
            last_term.value.kind is parse_retree.SymbolKind.END
        )

        if (
            len(node.union.uniates) != 1
            or not first_symbol_is_start
            or not last_symbol_is_end
        ):
            parts = [non_anchored_exception_message]
            if len(node.union.uniates) > 1:
                parts.append(
                    "Expected no alternation in the root group of "
                    "the pattern (only concatenation), but the pattern starts "
                    "with an alternation. You can not properly anchor "
                    "with an alternation."
                )

            if not first_symbol_is_start:
                parts.append(
                    f"Expected the first term of the pattern to be a start anchor, "
                    f"but it is not. Got: {_render_re_node(first_term)}"
                )

            if not last_symbol_is_end:
                parts.append(
                    f"Expected the last term of the pattern to be an end anchor, "
                    f"but it is not. Got: {_render_re_node(last_term)}"
                )

            raise NotImplementedError("\n\n".join(parts))

        check_for_formatted_value = _CheckForFormattedValue()
        check_for_formatted_value.visit(node)
        if check_for_formatted_value.has_formatted_value:
            raise AssertionError(
                f"The regex you want to transpile to the instructions of "
                f"a RegEx virtual machine contains a formatted value. "
                f"The formatted values can only be transpiled into code, "
                f"but can not be transpiled into instructions. "
                f"Please check your code logic. "
                f"The pattern was: {parse_retree.dump(node)}"
            )

        check_for_non_greedy_quantifiers = _CheckForNonGreedyQuantifiers()
        check_for_non_greedy_quantifiers.visit(node)
        if check_for_non_greedy_quantifiers.has_non_greedy_quantifiers:
            raise NotImplementedError(
                "(mristin, 2024-05-31): We did not implement the transpilation of "
                "non-greedy quantifiers to instructions of a RegEx virtual machine "
                "as this is more complex than the transpilation of the greedy ones. "
                "If you need this feature, please contact the developers."
            )

        assert len(node.union.uniates) == 1, (
            "Only concatenation expected at the root level since we must anchor "
            "at the start and at the end."
        )

        # region Optimize for arbitrary suffix

        # NOTE (mristin):
        # We optimize here for the pattern ``.*$`` as we can put an instruction ``Match``
        # just before the arbitrary suffix, and need not the match ``Any`` followed by
        # the ``End``.
        penultimate_term = (
            node.union.uniates[0].concatenants[-2]
            if len(node.union.uniates[0].concatenants) >= 2
            else None
        )

        assert (
            node.union.uniates[0].concatenants[0] is first_term
            and isinstance(first_term.value, parse_retree.Symbol)
            and first_term.value.kind is parse_retree.SymbolKind.START
        ), "Expected the first term to be an anchor at ``^``"

        assert (
            last_term is node.union.uniates[0].concatenants[-1]
            and isinstance(last_term.value, parse_retree.Symbol)
            and (last_term.value.kind is parse_retree.SymbolKind.END)
        ), "Expected the last term to be an anchor at ``$``"

        if (
            penultimate_term is not None
            and isinstance(penultimate_term.value, parse_retree.Symbol)
            and penultimate_term.value.kind is parse_retree.SymbolKind.DOT
            and penultimate_term.quantifier is not None
            and penultimate_term.quantifier.minimum == 0
            and penultimate_term.quantifier.maximum is None
        ):
            # NOTE (mristin):
            # We skip the start anchor as it is not transpiled to an instruction.
            concatenants = node.union.uniates[0].concatenants[1:-2]
        else:
            # NOTE (mristin):
            # We skip the start anchor as it is not transpiled to an instruction.
            concatenants = node.union.uniates[0].concatenants[1:]

        # endregion

        children = []  # type: List[_NodeOrLeaf]
        for concatenant in concatenants:
            child = self.transform(concatenant)
            children.append(child)

        children.append(_Leaf(instruction=InstructionMatch()))

        return _Node(
            re_node=node,
            children=children,
        )


def _recursively_convert_node_for_public(raw_node_or_leaf: _NodeOrLeaf) -> NodeOrLeaf:
    """
    Convert the post-processed "raw" node into a node for the public use.

    .. note::

        All post-processing needs to be performed *before* calling this function.
        This function only copy-converts the nodes into structures to be further
        used by the downstream clients. No post-processing is performed here.
    """
    if isinstance(raw_node_or_leaf, _Leaf):
        assert not isinstance(
            raw_node_or_leaf.instruction, _InstructionNoop
        ), "No no-op instructions expected in public"

        return Leaf(
            instruction=raw_node_or_leaf.instruction, label=raw_node_or_leaf.label
        )

    elif isinstance(raw_node_or_leaf, _Node):
        children = []  # type: List[NodeOrLeaf]
        for raw_child in raw_node_or_leaf.children:
            children.append(_recursively_convert_node_for_public(raw_child))

        return Node(re_node=raw_node_or_leaf.re_node, children=children)

    else:
        assert_never(raw_node_or_leaf)


def _linearize(node: _Node) -> List[_Leaf]:
    """Make recursively a linear list over all the leaves."""
    lst = []  # type: List[_Leaf]
    for child in node.children:
        if isinstance(child, _Leaf):
            lst.append(child)
        elif isinstance(child, _Node):
            lst.extend(_linearize(child))
        else:
            assert_never(child)

    return lst


def _relabel_in_place(root: _Node) -> None:
    """
    Re-assign labels according to the indices in a linearized sequence of instructions.

    We expect that the no-op instructions will be removed, so they are not assigned
    an index.
    """
    linearized_leaves = _linearize(root)

    # NOTE (mristin):
    # We index all the leaves except for no-op instructions which are going to be
    # eventually removed. The indices thus correspond to the instructions *after*
    # the no-op instructions are removed.

    next_index = 0

    # NOTE (mristin):
    # We map on the ``id(leaf)`` as the leaves are not hashable.
    leaf_to_index = dict()  # type: MutableMapping[int, int]

    for leaf in linearized_leaves:
        if isinstance(leaf.instruction, _InstructionNoop):
            continue

        leaf_to_index[id(leaf)] = next_index
        next_index += 1

    # NOTE (mristin):
    # This variable captures the mapping:
    # arbitrary labelling 🠒 labeling according to indices.
    #
    # Only the leaves indicated by new labels will finally have a label at the end.
    old_to_new_label = dict()  # type: MutableMapping[int, int]

    # NOTE (mristin):
    # We iterate in reverse over the leaves so that a single non-no-op instruction
    # can accumulate multiple labels for itself.

    leaf_after_noop = None  # type: Optional[_Leaf]
    for leaf in reversed(linearized_leaves):
        if isinstance(leaf.instruction, _InstructionNoop):
            if leaf.label is not None:
                assert leaf_after_noop is not None, (
                    "Expected at least one leaf *after* the no-op instruction. "
                    "Since the very last instruction must be a ``match`` instruction, "
                    "this must hold, so something obviously went wrong."
                )

                old_to_new_label[leaf.label] = leaf_to_index[id(leaf_after_noop)]
        else:
            if leaf.label is not None:
                old_to_new_label[leaf.label] = leaf_to_index[id(leaf)]

            leaf_after_noop = leaf

    new_label_set = set(old_to_new_label.values())

    for leaf in linearized_leaves:
        if isinstance(leaf.instruction, _InstructionNoop):
            leaf.label = None
            continue

        leaf_index = leaf_to_index[id(leaf)]

        if leaf_index in new_label_set:
            leaf.label = leaf_index
        else:
            leaf.label = None

        if isinstance(leaf.instruction, InstructionJump):
            leaf.instruction.target = old_to_new_label[leaf.instruction.target]

        elif isinstance(leaf.instruction, InstructionSplit):
            leaf.instruction.first_target = old_to_new_label[
                leaf.instruction.first_target
            ]

            leaf.instruction.second_target = old_to_new_label[
                leaf.instruction.second_target
            ]
        else:
            # NOTE (mristin):
            # Other instruction need not be adapted to the new labels.
            pass


def _remove_noop_in_place(node: _Node) -> None:
    """
    Remove recursively all no-op instructions in place.

    The no-op instructions are expected to have no labels attached to them.
    """
    new_children = []  # type: List[_NodeOrLeaf]
    for child in node.children:
        if isinstance(child, _Leaf):
            if isinstance(child.instruction, _InstructionNoop):
                assert child.label is None, (
                    f"Expected all no-op leaves to have their labels removed "
                    f"before calling {_remove_noop_in_place.__name__}"
                )
                continue

        elif isinstance(child, _Node):
            _remove_noop_in_place(child)

        else:
            assert_never(child)

        new_children.append(child)

    node.children = new_children


def translate(regex: parse_retree.Regex) -> Node:
    """Translate the regular expression into a program."""
    # NOTE (mristin):
    # We call the nodes "raw" here which still need to be post-processed.
    # The post-processing includes removal of no-ops, re-wiring of the labels *etc.*

    translator = _Translator()
    raw_root = translator.transform(regex)

    _relabel_in_place(raw_root)
    _remove_noop_in_place(raw_root)

    root = _recursively_convert_node_for_public(raw_node_or_leaf=raw_root)
    assert isinstance(root, Node), (
        f"Only node and no leaf expected at the root of the regular expression, "
        f"but got as root: {root}"
    )

    return root


def _determine_min_and_max_label(node_or_leaf: NodeOrLeaf) -> Optional[Tuple[int, int]]:
    """
    Iterate recursively over the nodes, and determine the extreme labels in the program.

    If no node contains a label, return ``None``.
    """
    if isinstance(node_or_leaf, Leaf):
        if node_or_leaf.label is not None:
            return node_or_leaf.label, node_or_leaf.label
        else:
            return None

    elif isinstance(node_or_leaf, Node):
        minimum = None  # type: Optional[int]
        maximum = None  # type: Optional[int]

        for child in node_or_leaf.children:
            maybe_child_min_max = _determine_min_and_max_label(node_or_leaf=child)
            if maybe_child_min_max is not None:
                child_min, child_max = maybe_child_min_max

                if minimum is not None:
                    minimum = min(minimum, child_min)
                else:
                    minimum = child_min

                if maximum is not None:
                    maximum = max(maximum, child_max)
                else:
                    maximum = child_max

        assert (minimum is not None and maximum is not None) or (
            minimum is None and maximum is None
        ), (
            f"Both minimum and maximum must be set or neither, "
            f"but got: {minimum=}, {maximum=}"
        )

        if minimum is not None:
            assert maximum is not None
            return minimum, maximum
        else:
            assert maximum is None
            return None
    else:
        assert_never(node_or_leaf)


@require(lambda indention: indention >= 0)
@require(lambda label_columns: label_columns >= 0)
def _write_recursively(
    node_or_leaf: NodeOrLeaf, indention: int, label_columns: int, writer: TextIO
) -> None:
    """Write recursively the node or leaf as a human-readable text to the ``writer``."""
    whitespace = "  " * indention

    if isinstance(node_or_leaf, Leaf):
        if label_columns == 0:
            label_prefix = ""
        else:
            if node_or_leaf.label is None:
                label_prefix = " " * (label_columns + 2)
            else:
                # noinspection PyStringFormat
                label_prefix = f"{{:0{label_columns}d}}: ".format(node_or_leaf.label)

        instruction_str: str

        if isinstance(node_or_leaf.instruction, InstructionChar):
            instruction_str = f"char {node_or_leaf.instruction.character!r}"

        elif isinstance(node_or_leaf.instruction, (InstructionSet, InstructionNotSet)):
            ranges_str = "".join(
                f"{rng.first}" if rng.first == rng.last else f"{rng.first}-{rng.last}"
                for rng in node_or_leaf.instruction.ranges
            )

            if isinstance(node_or_leaf.instruction, InstructionSet):
                instruction_str = f"set {ranges_str!r}"
            elif isinstance(node_or_leaf.instruction, InstructionNotSet):
                instruction_str = f"not-set {ranges_str!r}"
            else:
                assert_never(node_or_leaf.instruction)

        elif isinstance(node_or_leaf.instruction, InstructionAny):
            instruction_str = "any"

        elif isinstance(node_or_leaf.instruction, InstructionMatch):
            instruction_str = "match"

        elif isinstance(node_or_leaf.instruction, InstructionJump):
            instruction_str = f"jump {node_or_leaf.instruction.target}"

        elif isinstance(node_or_leaf.instruction, InstructionSplit):
            instruction_str = (
                f"split {node_or_leaf.instruction.first_target}, "
                f"{node_or_leaf.instruction.second_target}"
            )

        elif isinstance(node_or_leaf.instruction, InstructionEnd):
            instruction_str = "end"

        else:
            assert_never(node_or_leaf.instruction)

        writer.write(f"{label_prefix}{whitespace}{instruction_str}")

    elif isinstance(node_or_leaf, Node):
        if label_columns == 0:
            label_prefix = ""
        else:
            label_prefix = " " * (label_columns + 2)

        re_node_str = _render_re_node(node_or_leaf.re_node)
        if len(node_or_leaf.children) == 0:
            more_whitespace = "  " * (indention + 1)
            writer.write(
                f"{label_prefix}{whitespace}# {re_node_str}\n"
                f"{label_prefix}{whitespace}{{\n"
                f"{label_prefix}{more_whitespace}# Nothing\n"
                f"{label_prefix}{whitespace}}}"
            )
        elif len(node_or_leaf.children) == 1 and isinstance(
            node_or_leaf.children[0], Leaf
        ):
            if not isinstance(
                node_or_leaf.children[0].instruction,
                (InstructionSet, InstructionNotSet, InstructionAny, InstructionChar),
            ):
                writer.write(f"{label_prefix}{whitespace}# {re_node_str}\n")

            _write_recursively(
                node_or_leaf=node_or_leaf.children[0],
                indention=indention,
                label_columns=label_columns,
                writer=writer,
            )
        else:
            writer.write(
                f"{label_prefix}{whitespace}# {re_node_str}\n"
                f"{label_prefix}{whitespace}{{\n"
            )
            for child in node_or_leaf.children:
                _write_recursively(
                    node_or_leaf=child,
                    indention=indention + 1,
                    label_columns=label_columns,
                    writer=writer,
                )

                writer.write("\n")

            writer.write(f"{label_prefix}{whitespace}}}")

    else:
        assert_never(node_or_leaf)


def dump(program: NodeOrLeaf) -> str:
    """Write out the program as a readable nested list of instructions."""
    maybe_min_max_label = _determine_min_and_max_label(node_or_leaf=program)

    if maybe_min_max_label is None:
        label_columns = 0
    else:
        min_label, max_label = maybe_min_max_label
        label_columns = max(len(str(min_label)), len(str(max_label)))

    writer = io.StringIO()
    _write_recursively(
        node_or_leaf=program, indention=0, label_columns=label_columns, writer=writer
    )

    return writer.getvalue()



"""Generate the Golang code for JSON-ization based on the intermediate representation."""

import io
from typing import Tuple, Optional, List, Union

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
    assert_union_without_excluded,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    description as golang_description,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# region De-serialization


def _generate_bool_from_jsonable() -> Stripped:
    """Generate the function to decode a ``bool`` from a JSON-able."""
    return Stripped(
        f"""\
// Parse `jsonable` as a boolean, or return an error.
func boolFromJsonable(
{I}jsonable interface{{}},
) (result bool, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected a boolean, but got null",
{II})
{II}return
{I}}}

{I}var ok bool
{I}result, ok = jsonable.(bool)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf("Expected a boolean, but got %T", jsonable),
{II})

{II}return
{I}}}

{I}return
}}"""
    )


def _generate_int64_from_jsonable() -> Stripped:
    """Generate the function to decode an ``int64`` from a JSON-able."""
    return Stripped(
        f"""\
// Parse `jsonable` as a 64-bit integer, or return an error.
func int64FromJsonable(
{I}jsonable interface{{}},
) (result int64, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected an integer number, but got null",
{II})
{II}return
{I}}}

{I}f, ok := jsonable.(float64)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an integer number, but got %T",
{IIII}jsonable,
{III}),
{II})
{II}return
{I}}}

{I}if math.IsNaN(f) {{
{II}err = newDeserializationError(
{III}"Expected an integer number, but got a NaN",
{II})
{II}return
{I}}}

{I}if math.IsInf(f, 0) {{
{II}err = newDeserializationError(
{III}"Expected an integer number, but got an infinity",
{II})
{II}return
{I}}}

{I}if f != math.Trunc(f) {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an integer number, but got a non-integer: %v",
{IIII}f,
{III}),
{II})
{II}return
{I}}}

{I}result = int64(f)
{I}if f != float64(result) {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an integer number fitting into int64, but got: %v",
{IIII}jsonable,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_float64_from_jsonable() -> Stripped:
    """Generate the function to decode a ``float`` from a JSON-able."""
    return Stripped(
        f"""\
// Parse `jsonable` as a 64-bit float, or return an error.
func float64FromJsonable(
{I}jsonable interface{{}},
) (result float64, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected a number, but got null",
{II})
{II}return
{I}}}

{I}var ok bool
{I}result, ok = jsonable.(float64)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a number, but got %T",
{IIII}jsonable,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_string_from_jsonable() -> Stripped:
    """Generate the function to decode a ``string`` from a JSON-able."""
    return Stripped(
        f"""\
// Parse `jsonable` as a string, or return an error.
func stringFromJsonable(
{I}jsonable interface{{}},
) (result string, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected a string, but got null",
{II})
{II}return
{I}}}

{I}var ok bool
{I}result, ok = jsonable.(string)
{I}if ok {{
{II}return
{I}}} else {{
{II}err = newDeserializationError(
{III}fmt.Sprintf("Expected a string, but got %T", jsonable),
{II})
{II}return
{I}}}
}}"""
    )


def _generate_bytes_from_jsonable() -> Stripped:
    """Generate the function to decode ``[]byte`` from a JSON-able."""
    return Stripped(
        f"""\
// Parse `jsonable` as a byte array, or return an error.
func bytesFromJsonable(
{I}jsonable interface{{}},
) (result []byte, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected a base64-encoded string, but got null",
{II})
{II}return
{I}}}

{I}text, ok := jsonable.(string)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a base64-encoded string, but got %T",
{IIII}jsonable,
{III}),
{II})
{II}return
{I}}}

{I}var decodingErr error
{I}result, decodingErr = b64.StdEncoding.DecodeString(text)
{I}if decodingErr != nil {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"String could not be decoded as base64: %s",
{IIII}decodingErr.Error(),
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_enumeration_from_jsonable(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the deserialization method for an enumeration."""
    enum_name = golang_naming.enum_name(identifier=enumeration.name)

    function_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_jsonable")
    )

    enum_from_str = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    return Stripped(
        f"""\
// Parse `jsonable` as a literal of [aastypes.{enum_name}],
// or return an error.
func {function_name}(
{I}jsonable interface{{}},
) (result aastypes.{enum_name}, err error) {{
{I}if jsonable == nil {{
{II}err = newDeserializationError(
{III}"Expected a string representation of {enum_name}, " +
{III}"but got null",
{II})
{II}return
{I}}}

{I}text, ok := jsonable.(string)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a string representation of {enum_name}, " +
{IIII}"but got %T",
{IIII}jsonable,
{III}),
{II})
{II}return
{I}}}

{I}result, ok = aasstringification.{enum_from_str}(text)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a string representation of {enum_name}, " +
{IIII}"but got %v",
{IIII}text,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


# fmt: off
@require(
    lambda cls:
    len(cls.concrete_descendants) > 0,
    "A class that needs dispatch must have more than one concrete descendant"
)
# fmt: on
def _generate_class_from_map(cls: intermediate.ClassUnion) -> Stripped:
    """Generate a mapping model type 🠒 de-serialization from map."""
    model_types_dispatch_names = []  # type: List[Tuple[str, str]]

    for descendant in cls.concrete_descendants:
        model_types_dispatch_names.append(
            (
                naming.json_model_type(descendant.name),
                golang_naming.private_function_name(
                    Identifier(f"{descendant.name}_from_map_without_dispatch")
                ),
            )
        )

    if isinstance(cls, intermediate.ConcreteClass):
        model_types_dispatch_names.append(
            (
                naming.json_model_type(cls.name),
                golang_naming.private_function_name(
                    Identifier(f"{cls.name}_from_map_without_dispatch")
                ),
            )
        )

    case_blocks = []  # type: List[Stripped]
    for model_type, dispatch_name in model_types_dispatch_names:
        model_type_literal = golang_common.string_literal(model_type)
        case_blocks.append(
            Stripped(
                f"""\
case {model_type_literal}:
{I}result, err = {dispatch_name}(m)"""
            )
        )

    interface_name = golang_naming.interface_name(cls.name)

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}err = newDeserializationError(
{II}fmt.Sprintf(
{III}"Unexpected model type " +
{III}"for {interface_name}: %s",
{III}modelType,
{II}),
{I})"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    function_name = golang_naming.private_function_name(
        Identifier(f"{cls.name}_from_map")
    )

    return Stripped(
        f"""\
// De-serialize an instance of [aastypes.{interface_name}]
// from a map by dispatching to the concrete `*FromMapWithoutDispatch` function.
func {function_name}(
{I}m map[string]interface{{}},
) (
{I}result aastypes.{interface_name},
{I}err error,
) {{
{I}var modelTypeAny interface{{}}
{I}var ok bool
{I}modelTypeAny, ok = m["modelType"];
{I}if !ok {{
{II}err = newDeserializationError(
{III}"The required property modelType is missing",
{II})
{II}return
{I}}}

{I}var modelType string
{I}modelType, ok = modelTypeAny.(string)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected the property modelType to be a string, " +
{IIII}"but got %T",
{IIII}modelTypeAny,
{III}),
{II})
{II}return
{I}}}

{I}switch modelType {{
{I}{indent_but_first_line(case_blocks_joined, I)}
{I}}}

{I}return
}}"""
    )


def _generate_class_from_jsonable(cls: intermediate.ClassUnion) -> Stripped:
    """Generate the de-serialization function for a class that involves a dispatch."""
    function_name = golang_naming.function_name(Identifier(f"{cls.name}_from_jsonable"))

    interface_name = golang_naming.interface_name(cls.name)

    blocks = [
        Stripped(
            f"""\
if jsonable == nil {{
{I}err = newDeserializationError(
{II}"Expected a JSON object, but got null",
{I})
{I}return
}}"""
        ),
        Stripped(
            f"""\
m, ok := jsonable.(map[string]interface{{}})
if !ok {{
{I}err = newDeserializationError(
{II}fmt.Sprintf(
{III}"Expected a JSON object, but got %T",
{III}jsonable,
{II}),
{I})
{I}return
}}"""
        ),
    ]  # type: List[Stripped]

    if len(cls.concrete_descendants) == 0:
        assert not isinstance(cls, intermediate.AbstractClass), (
            "We can not parse abstract classes without any concrete descendants "
            "as we do not know the concrete structure of the map."
        )

        from_map_name = golang_naming.private_function_name(
            Identifier(f"{cls.name}_from_map_without_dispatch")
        )
    else:
        from_map_name = golang_naming.private_function_name(
            Identifier(f"{cls.name}_from_map")
        )

    blocks.append(Stripped(f"result, err = {from_map_name}(m)"))
    blocks.append(Stripped("return"))

    body = Stripped("\n\n".join(blocks))

    return Stripped(
        f"""\
// Parse `jsonable` as an instance of [aastypes.{interface_name}],
// or return an error.
func {function_name}(
{I}jsonable interface{{}},
) (
{I}result aastypes.{interface_name},
{I}err error,
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


_PARSE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "boolFromJsonable",
    intermediate.PrimitiveType.INT: "int64FromJsonable",
    intermediate.PrimitiveType.FLOAT: "float64FromJsonable",
    intermediate.PrimitiveType.STR: "stringFromJsonable",
    intermediate.PrimitiveType.BYTEARRAY: "bytesFromJsonable",
}
assert all(
    literal in _PARSE_FUNCTION_BY_PRIMITIVE_TYPE
    for literal in intermediate.PrimitiveType
)


def _determine_parse_function_for_atomic_value(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Stripped:
    """Determine the parse function for deserializing an atomic non-optional value."""
    function_name: str

    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            function_name = golang_naming.function_name(
                Identifier(f"{our_type.name}_from_jsonable")
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[our_type.constrainee]

        elif isinstance(
            our_type,
            (
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            function_name = golang_naming.function_name(
                Identifier(f"{our_type.name}_from_jsonable")
            )

        else:
            assert_never(our_type)
    else:
        assert_never(type_annotation)

    return Stripped(function_name)


def _generate_deserialization_switch_statement(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate the switch statement for de-serialization of the properties.

    This statement is expected to be run in a ``for k, v := range m`` loop.
    We switch on ``k`` and the value of the property is given as ``v``. The ``jsonable``
    is expected to be of type ``map[string]interface{}``.

    The resulting struct we parse into is ``result``, a pointer to the struct
    corresponding to ``cls``. Whenever we encounter a property, we have to set
    the corresponding boolean ``found*`` to ``True``.

    This function was originally part of
    ``_generate_concrete_class_from_map_without_dispatch``, but we refactored it out
    since it was too much to read. Best if your read both functions in two vertical
    editor panes.

    If the serialization requires model type, we also expect that the variable
    ``foundModelType`` have been initialized as a ``bool`` in the preceding generated
    code. The generated case block will parse and set it in case that the model type
    is correctly specified. This will be performed even though the code might have
    had to parse model type before for the dispatch. We decided to double-check to cover
    the case where a dispatch is *unnecessary* (*e.g.*, the caller knows the expected
    runtime type), but the model type might still be invalid in the input. Hence, when
    the dispatch is *necessary*, the model type JSON property will be parsed twice,
    which is a cost we currently find acceptable.
    """
    case_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        optional = isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)

        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))

        json_prop_literal = golang_common.string_literal(
            naming.json_property(prop.name)
        )

        case_body: Stripped

        primitive_type = intermediate.try_primitive_type(type_anno)

        # NOTE (mristin, 2023-05-12):
        # We handle this case of the atomic value separately from the others as
        # we model the optional values with pointers, so we first have to parse into
        # a value, and then pass a pointer to that value to the property of
        # the instance.
        if optional and (
            (
                primitive_type is not None
                and primitive_type is not intermediate.PrimitiveType.BYTEARRAY
            )
            or (
                isinstance(type_anno, intermediate.OurTypeAnnotation)
                and isinstance(type_anno.our_type, intermediate.Enumeration)
            )
        ):
            assert isinstance(
                type_anno,
                (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
            )
            parse_function = _determine_parse_function_for_atomic_value(type_anno)

            value_type = golang_common.generate_type(
                type_annotation=type_anno, types_package=Identifier("aastypes")
            )

            case_body = Stripped(
                f"""\
var parsed {value_type}
parsed, err = {parse_function}(
{I}v,
)
if err != nil {{
{I}if deseriaErr, ok := err.(*DeserializationError); ok {{
{II}deseriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {json_prop_literal},
{III}}},
{II})
{I}}}
{I}return
}}
{prop_var} = &parsed"""
            )

        elif isinstance(
            type_anno,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ):
            parse_function = _determine_parse_function_for_atomic_value(type_anno)

            case_body = Stripped(
                f"""\
{prop_var}, err = {parse_function}(
{I}v,
)
if err != nil {{
{I}if deseriaErr, ok := err.(*DeserializationError); ok {{
{II}deseriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {json_prop_literal},
{III}}},
{II})
{I}}}
{I}return
}}"""
            )

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin, 2023-03-29): We expect only lists of classes "
                f"at the moment, but you specified {type_anno}. "
                f"Please contact the developers if you need this feature."
            )

            parse_function = _determine_parse_function_for_atomic_value(type_anno.items)

            array_type = golang_common.generate_type(
                type_annotation=type_anno, types_package=Identifier("aastypes")
            )

            item_type = golang_common.generate_type(
                type_annotation=type_anno.items, types_package=Identifier("aastypes")
            )

            case_body = Stripped(
                f"""\
jsonableArray, ok := v.([]interface{{}})
if !ok {{
{I}deseriaErr := newDeserializationError(
{II}fmt.Sprintf(
{III}"Expected an array, but got %T",
{III}v,
{II}),
{I})

{I}deseriaErr.Path.PrependName(
{II}&aasreporting.NameSegment{{
{III}Name: {json_prop_literal},
{II}}},
{I})

{I}err = deseriaErr

{I}return
}}

array := make(
{I}{array_type},
{I}len(jsonableArray),
)
for i, itemJsonable := range jsonableArray {{
{I}var item {item_type}
{I}item, err = {parse_function}(
{II}itemJsonable,
{I})
{I}if err != nil {{
{II}if deseriaErr, ok := err.(*DeserializationError); ok {{
{III}deseriaErr.Path.PrependIndex(
{IIII}&aasreporting.IndexSegment{{
{IIIII}Index: i,
{IIII}}},
{III})

{III}deseriaErr.Path.PrependName(
{IIII}&aasreporting.NameSegment{{
{IIIII}Name: {json_prop_literal},
{IIII}}},
{III})
{II}}}

{II}return
{I}}}

{I}array[i] = item
}}
{prop_var} = array"""
            )

        else:
            assert_never(type_anno)

        if not optional:
            found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))

            # NOTE (mristin, 2023-04-09):
            # Appending once is OK for time complexity. If you append more, please
            # refactor into a list and join.
            case_body = Stripped(
                f"""\
{case_body}
{found_var} = true"""
            )

        case_blocks.append(
            Stripped(
                f"""\
case {json_prop_literal}:
{I}{indent_but_first_line(case_body, I)}"""
            )
        )

    if cls.serialization.with_model_type:
        # NOTE (mristin):
        # We explicitly check for the model type if the meta-model instructs us to.
        #
        # This is crucial for the fix of the problem discovered in:
        # https://github.com/aas-core-works/aas-core3.0-python/issues/32
        model_type = naming.json_model_type(cls.name)
        model_type_case_body = Stripped(
            f"""\
var modelType string
modelType, err = stringFromJsonable(
{I}v,
)
if err != nil {{
{I}if deseriaErr, ok := err.(*DeserializationError); ok {{
{II}deseriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: "modelType",
{III}}},
{II})
{I}}}
{I}return
}}

if modelType != "{model_type}" {{
{I}deseriaErr := newDeserializationError(
{II}fmt.Sprintf(
{III}"Expected the model type '{model_type}', but got %v",
{III}v,
{II}),
{I})

{I}deseriaErr.Path.PrependName(
{II}&aasreporting.NameSegment{{
{III}Name: "modelType",
{II}}},
{I})

{I}err = deseriaErr
{I}return
}}

foundModelType = true"""
        )
        case_blocks.append(
            Stripped(
                f"""\
case "modelType":
{I}{indent_but_first_line(model_type_case_body, I)}"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}err = newDeserializationError(
{II}fmt.Sprintf(
{III}"Unexpected property: %s",
{III}k,
{II}),
{I})
{I}return"""
        )
    )

    case_blocks_joined = "\n\n".join(case_blocks)

    return Stripped(
        f"""\
switch k {{
{case_blocks_joined}
}}"""
    )


def _generate_concrete_class_from_map_without_dispatch(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate the deserialization function for a concrete class from a map.

    This function performs no dispatch. If it de-serializes a concrete class with
    concrete descendants, we have to provide a different name. Otherwise, it would
    shadow the name for the dispatch function.
    """
    # fmt: off
    assert (
            sorted(
                (arg.name, str(arg.type_annotation))
                for arg in cls.constructor.arguments
            ) == sorted(
                (prop.name, str(prop.type_annotation))
                for prop in cls.properties
            )
    ), (
        "(mristin, 2023-04-07) We assume that the properties and constructor arguments "
        "are identical at this point. If this is not the case, we have to re-write the "
        "logic substantially! Please contact the developers if you see this."
    )
    # fmt: on

    blocks = []  # type: List[Stripped]

    # region Initialize

    prop_var_initializations = []  # type: List[Stripped]
    for prop in cls.properties:
        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))

        prop_var_type = golang_common.generate_type(
            type_annotation=prop.type_annotation, types_package=Identifier("aastypes")
        )

        prop_var_initializations.append(Stripped(f"var {prop_var} {prop_var_type}"))

    if len(prop_var_initializations) > 0:
        blocks.append(Stripped("\n".join(prop_var_initializations)))

    found_var_initializations = []  # type: List[Stripped]
    for prop in cls.properties:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))

        found_var_initializations.append(Stripped(f"{found_var} := false"))

    if len(found_var_initializations) > 0:
        blocks.append(Stripped("\n".join(found_var_initializations)))

    # endregion

    # NOTE (mristin):
    # We explicitly check for the model type if the meta-model instructs us to.
    #
    # This is crucial for the fix of the problem discovered in:
    # https://github.com/aas-core-works/aas-core3.0-python/issues/32
    if cls.serialization.with_model_type:
        blocks.append(Stripped("var foundModelType bool"))

    # region Switch on property name

    switch_statement = _generate_deserialization_switch_statement(cls=cls)

    # endregion

    blocks.append(
        Stripped(
            f"""\
for k, v := range m {{
{I}{indent_but_first_line(switch_statement, I)}
}}"""
        )
    )

    # region Check required properties

    for prop in cls.properties:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))

        message_literal = golang_common.string_literal(
            f"The required property {naming.json_property(prop.name)!r} is missing"
        )

        blocks.append(
            Stripped(
                f"""\
if !{found_var} {{
{I}err = newDeserializationError(
{II}{message_literal},
{I})
{I}return
}}"""
            )
        )

    # endregion

    if cls.serialization.with_model_type:
        # NOTE (mristin):
        # We explicitly check for the model type if the meta-model instructs us to.
        #
        # This is crucial for the fix of the problem discovered in:
        # https://github.com/aas-core-works/aas-core3.0-python/issues/32
        blocks.append(
            Stripped(
                f"""\
if !foundModelType {{
{I}err = newDeserializationError(
{II}"The required property modelType is missing",
{I})
{I}return
}}"""
            )
        )

    constructing_statements = []  # type: List[Stripped]

    constructor_arguments = [
        golang_naming.variable_name(Identifier(f"the_{arg.name}"))
        for arg in cls.constructor.arguments
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation)
    ]  # type: List[Stripped]

    new_function = golang_naming.function_name(Identifier(f"new_{cls.name}"))

    if len(constructor_arguments) > 0:
        constructor_arguments_joined = "\n".join(
            f"{arg}," for arg in constructor_arguments
        )

        constructing_statements.append(
            Stripped(
                f"""\
result = aastypes.{new_function}(
{I}{indent_but_first_line(constructor_arguments_joined, I)}
)"""
            )
        )
    else:
        constructing_statements.append(Stripped(f"result = aastypes.{new_function}()"))

    for arg in cls.constructor.arguments:
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        setter_name = golang_naming.setter_name(arg.name)
        prop_var = golang_naming.variable_name(Identifier(f"the_{arg.name}"))

        constructing_statements.append(
            Stripped(
                f"""\
result.{setter_name}(
{I}{prop_var},
)"""
            )
        )

    blocks.append(Stripped("\n".join(constructing_statements)))

    blocks.append(Stripped("return"))

    body = Stripped("\n\n".join(blocks))

    interface_name = golang_naming.interface_name(cls.name)

    function_name = golang_naming.private_function_name(
        Identifier(f"{cls.name}_from_map_without_dispatch")
    )

    documentation_blocks = [
        Stripped(
            f"""\
Parse [aastypes.{interface_name}] from a map,
or return an error, if any."""
        )
    ]  # type: List[Stripped]

    if len(cls.concrete_descendants) > 0:
        function_name_from_jsonable = golang_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        documentation_blocks.append(
            Stripped(
                f"""\
This function performs no dispatch! It is used to parse the properties
as-are, and already assumes the exact model type. Usually, this function
is called from within a from-jsonable or from-map function, and you never
call it directly. If you want to de-serialize an instance of
[aastypes.{interface_name}], call
[{function_name_from_jsonable}]."""
            )
        )

    documentation_comment = golang_description.documentation_comment(
        Stripped("\n\n".join(documentation_blocks))
    )

    return Stripped(
        f"""\
{documentation_comment}
func {function_name}(
{I}m map[string]interface{{}},
) (
{I}result aastypes.{interface_name},
{I}err error,
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


# endregion

# region Serialization


def _generate_int64_to_jsonable() -> Stripped:
    """Generate the function to encode an ``int64`` to a JSON-able."""
    return Stripped(
        f"""\
// Try to cast `that` to a float64, or return an error.
func int64ToJsonable(
{I}that int64,
) (result float64, err error) {{
{I}if that > 9007199254740991 || that < -9007199254740991 {{
{II}err = newSerializationError(
{III}fmt.Sprintf(
{IIII}"64-bit integer can not be represented as 64-bit float in JSON: %v",
{IIII}that,
{III}),
{II})
{II}return
{I}}}

{I}result = float64(that);
{I}return
}}"""
    )


def _generate_bytes_to_jsonable() -> Stripped:
    """Generate the function to encode ``[]byte`` to a string."""
    return Stripped(
        f"""\
// Encode `bytes` to a base64 string.
func bytesToJsonable(
{I}bytes []byte,
) (result string, err error) {{
{I}if bytes == nil {{
{II}err = newSerializationError(
{III}"Expected an array of bytes, but got nil",
{II})
{II}return
{I}}}

{I}result = b64.StdEncoding.EncodeToString(
{II}bytes,
{I})
{I}return
}}"""
    )


def _generate_enumeration_to_jsonable(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the serialization method for an enumeration."""
    enum_name = golang_naming.enum_name(identifier=enumeration.name)

    function_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_to_jsonable")
    )

    enum_to_str = golang_naming.function_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    return Stripped(
        f"""\
// Serialize `that` to a string, or return an error.
func {function_name}(
{I}that aastypes.{enum_name},
) (result string, err error) {{
{I}var ok bool
{I}result, ok = aasstringification.{enum_to_str}(
{II}that,
{I})
{I}if !ok {{
{II}err = newSerializationError(
{III}fmt.Sprintf(
{IIII}"Got an invalid literal of {enum_name}: %v",
{IIII}that,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


TypeAnnotationExceptList = Union[
    intermediate.PrimitiveTypeAnnotation,
    intermediate.OurTypeAnnotation,
    intermediate.OptionalTypeAnnotation,
]

assert_union_without_excluded(
    original_union=intermediate.TypeAnnotationUnion,
    subset_union=TypeAnnotationExceptList,
    excluded=[intermediate.ListTypeAnnotation],
)


def _generate_expression_to_serialize_atomic_value(
    access_expression: str, type_annotation: TypeAnnotationExceptList
) -> Tuple[Stripped, bool]:
    """
    Generate the snippet to serialize the ``access_expression``.

    The ``access_expression`` is for example a name or a property access.
    The caller is expected to have already generated the code which checks that
    ``access_expression`` is not nil.

    Return (expression, True if there needs to be error checking)
    """
    type_anno = intermediate.beneath_optional(type_annotation)
    assert isinstance(
        type_anno,
        (
            intermediate.PrimitiveTypeAnnotation,
            intermediate.OurTypeAnnotation,
        ),
    )

    optional = isinstance(type_annotation, intermediate.OptionalTypeAnnotation)

    # NOTE (mristin, 2023-04-12):
    # The following types are handled as primitive values in Golang, so we have
    # to handle them as such. They are primitive values if non-nullable, and referenced
    # if nullable, since we model optional primitive properties as pointers in
    # Golang.
    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.our_type,
            (intermediate.Enumeration, intermediate.ConstrainedPrimitive),
        )
    ):
        primitive_type = intermediate.try_primitive_type(type_anno)

        if primitive_type is intermediate.PrimitiveType.INT:
            if not optional:
                return (
                    Stripped(
                        f"""\
int64ToJsonable(
{I}{access_expression},
)"""
                    ),
                    True,
                )
            else:
                return (
                    Stripped(
                        f"""\
int64ToJsonable(
{I}*({access_expression}),
)"""
                    ),
                    True,
                )

        elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
            return (
                Stripped(
                    f"""\
bytesToJsonable(
{I}{access_expression},
)"""
                ),
                True,
            )

        elif isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.our_type, intermediate.Enumeration
        ):
            enumeration = type_anno.our_type
            enum_to_jsonable = golang_naming.function_name(
                Identifier(f"{enumeration.name}_to_jsonable")
            )

            if not optional:
                return (
                    Stripped(
                        f"""\
{enum_to_jsonable}(
{I}{access_expression},
)"""
                    ),
                    True,
                )
            else:
                return (
                    Stripped(
                        f"""\
{enum_to_jsonable}(
{I}*({access_expression}),
)"""
                    ),
                    True,
                )
        else:
            if optional:
                return Stripped(f"*{access_expression}"), False
            else:
                return Stripped(access_expression), False

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            raise AssertionError(f"Should have been handled before: {type_anno=}")

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            raise AssertionError(f"Should have been handled before: {type_anno=}")

        elif isinstance(
            our_type,
            (
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            return (
                Stripped(
                    f"""\
ToJsonable(
{I}{access_expression},
)"""
                ),
                True,
            )

        else:
            assert_never(our_type)
    else:
        assert_never(type_anno)


def _generate_cls_to_map(cls: intermediate.ConcreteClass) -> Stripped:
    """
    Generate the function to serialize class to a JSON-able map.

    The generated function will perform no dispatching.
    """
    blocks = [Stripped("result = make(map[string]interface{})")]  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        getter_name = golang_naming.getter_name(prop.name)

        prop_literal = golang_common.string_literal(
            f"{golang_naming.property_name(prop.name)}()"
        )

        json_prop_literal = golang_common.string_literal(
            naming.json_property(prop.name)
        )

        prop_jsonable_var = golang_naming.variable_name(
            Identifier(f"jsonable_{prop.name}")
        )

        block: Stripped

        if isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items,
                intermediate.OurTypeAnnotation,
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                "(mristin, 2023-04-12): We expect only lists of our classes. "
                "Other lists are not handled yet. Please contact the developers."
            )

            statements = [
                Stripped(
                    f"""\
{prop_jsonable_var} := make(
{I}[]interface{{}},
{I}len(that.{getter_name}()),
)"""
                )
            ]  # type: List[Stripped]

            # fmt: off
            serialize_expr, needs_error_checking = (
                _generate_expression_to_serialize_atomic_value(
                    access_expression="v", type_annotation=type_anno.items
                )
            )
            # fmt: on

            if needs_error_checking:
                statements.append(
                    Stripped(
                        f"""\
for i, v := range that.{getter_name}() {{
{I}var jsonable interface{{}}
{I}jsonable, err = {indent_but_first_line(serialize_expr, I)}
{I}if err != nil {{
{II}if seriaErr, ok := err.(*SerializationError); ok {{
{III}seriaErr.Path.PrependIndex(
{IIII}&aasreporting.IndexSegment{{
{IIIII}Index: i,
{IIII}}},
{III})

{III}seriaErr.Path.PrependName(
{IIII}&aasreporting.NameSegment{{
{IIIII}Name: {prop_literal},
{IIII}}},
{III})
{II}}}

{II}return
{I}}}
{I}{prop_jsonable_var}[i] = jsonable
}}"""
                    )
                )
            else:
                statements.append(
                    Stripped(
                        f"""\
for _, v := range that.{getter_name}() {{
{I}{prop_jsonable_var} = append(
{II}{prop_jsonable_var},
{II}{indent_but_first_line(serialize_expr, II)}
{I})
}}"""
                    )
                )

            statements.append(
                Stripped(
                    f"""\
result[{json_prop_literal}] = {prop_jsonable_var}"""
                )
            )

            block = Stripped("\n".join(statements))
        else:
            assert not isinstance(prop.type_annotation, intermediate.ListTypeAnnotation)

            # fmt: off
            serialize_expr, needs_error_checking = (
                _generate_expression_to_serialize_atomic_value(
                    access_expression=f"that.{getter_name}()",
                    type_annotation=prop.type_annotation
                )
            )
            # fmt: on

            if needs_error_checking:
                block = Stripped(
                    f"""\
var {prop_jsonable_var} interface{{}}
{prop_jsonable_var}, err = {serialize_expr}
if err != nil {{
{I}if seriaErr, ok := err.(*SerializationError); ok {{
{II}seriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {prop_literal},
{III}}},
{II})
{I}}}

{I}return
}}
result[{json_prop_literal}] = {prop_jsonable_var}"""
                )
            else:
                block = Stripped(
                    f"""\
result[{json_prop_literal}] = {serialize_expr}"""
                )

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            block = Stripped(
                f"""\
if that.{getter_name}() != nil {{
{I}{indent_but_first_line(block, I)}
}}"""
            )

        blocks.append(block)

    if cls.serialization.with_model_type:
        model_type_literal = golang_common.string_literal(
            naming.json_model_type(cls.name)
        )
        blocks.append(Stripped(f'result["modelType"] = {model_type_literal}'))

    blocks.append(Stripped("return"))

    body = "\n\n".join(blocks)

    function_name = golang_naming.private_function_name(
        Identifier(f"{cls.name}_to_map")
    )

    interface_name = golang_naming.interface_name(cls.name)

    to_jsonable = golang_naming.function_name(Identifier("to_jsonable"))

    return Stripped(
        f"""\
// Serialize [aastypes.{interface_name}] as a JSON-able map.
//
// This function performs no dispatch! It is only used to serialize
// the properties. If you want to serialize an instance of
// [aastypes.{interface_name}] with proper dispatch, call
// [{to_jsonable}].
func {function_name}(
{I}that aastypes.{interface_name},
) (result map[string]interface{{}}, err error) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_to_jsonable(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the main entry point for the serialization."""
    case_blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        interface_name = golang_naming.interface_name(cls.name)

        to_map = golang_naming.private_function_name(Identifier(f"{cls.name}_to_map"))

        case_blocks.append(
            Stripped(
                f"""\
case aastypes.{literal}:
{I}result, err = {to_map}(
{II}that.(aastypes.{interface_name}),
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}err = newSerializationError(
{II}fmt.Sprintf(
{III}"Unexpected model type literal: %v",
{III}that.ModelType(),
{II}),
{I})"""
        )
    )

    switch_body = Stripped("\n".join(case_blocks))
    model_type_getter = golang_naming.getter_name(Identifier("model_type"))
    switch_statement = Stripped(
        f"""\
switch that.{model_type_getter}() {{
{switch_body}
}}"""
    )

    return Stripped(
        f"""\
// Serialize “that“ instance to a JSON-able representation.
//
// Return a structure which can be readily converted to JSON,
// or an error if some value could not be converted.
func ToJsonable(
{I}that aastypes.IClass,
) (result map[string]interface{{}}, err error) {{
{I}{indent_but_first_line(switch_statement, I)}
{I}return
}}"""
    )


# endregion

# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    repo_url: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code for the general de/serialization."""
    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    aasreporting_url_literal = golang_common.string_literal(f"{repo_url}/reporting")

    aasstringification_url_literal = golang_common.string_literal(
        f"{repo_url}/stringification"
    )

    blocks = [
        Stripped(
            """\
// Package jsonization de/serializes model instances to and from JSON.
//
// We can not use one-pass deserialization for JSON since the object
// properties do not have fixed order, and hence we can not read
// `modelType` property ahead of the remaining properties.
//
// To de-serialize, call one of the `*FromJsonable` functions.
//
// To serialize, call [ToJsonable] function.
package jsonization"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"math"
{I}b64 "encoding/base64"
{I}aasreporting {aasreporting_url_literal}
{I}aasstringification {aasstringification_url_literal}
{I}aastypes {aastypes_url_literal}
)"""
        ),
        Stripped("// region De-serialization"),
        Stripped(
            f"""\
// Represent an error during the de-serialization.
//
// Implements `error`.
type DeserializationError struct{{
{I}Path *aasreporting.Path
{I}Message string
}}"""
        ),
        Stripped(
            f"""\
func newDeserializationError(message string) *DeserializationError {{
{I}return &DeserializationError{{
{II}Path: &aasreporting.Path{{}},
{II}Message: message,
{I}}}
}}"""
        ),
        Stripped(
            f"""\
func (de *DeserializationError) Error() string {{
{I}return fmt.Sprintf(
{II}"%s: %s",
{II}de.PathString(),
{II}de.Message,
{I})
}}"""
        ),
        Stripped(
            f"""\
// Render the path as a string.
func (de *DeserializationError) PathString() string {{
{I}return aasreporting.ToJSONPath(de.Path)
}}"""
        ),
        _generate_bool_from_jsonable(),
        _generate_int64_from_jsonable(),
        _generate_float64_from_jsonable(),
        _generate_string_from_jsonable(),
        _generate_bytes_from_jsonable(),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_enumeration_from_jsonable(enumeration=our_type))
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            pass
        elif isinstance(our_type, intermediate.AbstractClass):
            blocks.append(_generate_class_from_jsonable(cls=our_type))
        elif isinstance(our_type, intermediate.ConcreteClass):
            blocks.append(_generate_class_from_jsonable(cls=our_type))

            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Jsonization/{our_type.name}_from_map.go"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The jsonization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue
            else:
                blocks.append(
                    _generate_concrete_class_from_map_without_dispatch(cls=our_type)
                )
        else:
            assert_never(our_type)

    # NOTE (mristin, 2023-04-12):
    # We add all the dispatch mappings at the end as the functions might not have been
    # defined yet.
    for cls in symbol_table.classes:
        if isinstance(cls, intermediate.AbstractClass):
            blocks.append(_generate_class_from_map(cls=cls))
        elif isinstance(cls, intermediate.ConcreteClass):
            if len(cls.concrete_descendants) > 0:
                blocks.append(_generate_class_from_map(cls=cls))
        else:
            assert_never(cls)

    blocks.append(Stripped("// endregion"))

    blocks.append(Stripped("// region Serialization"))

    blocks.extend(
        [
            Stripped(
                f"""\
// Represent an error during the serialization.
//
// Implements `error`.
type SerializationError struct{{
{I}Path *aasreporting.Path
{I}Message string
}}"""
            ),
            Stripped(
                f"""\
func newSerializationError(message string) *SerializationError {{
{I}return &SerializationError{{
{II}Path: &aasreporting.Path{{}},
{II}Message: message,
{I}}}
}}"""
            ),
            Stripped(
                f"""\
func (se *SerializationError) Error() string {{
{I}return fmt.Sprintf(
{II}"%s: %s",
{II}se.PathString(),
{II}se.Message,
{I})
}}"""
            ),
            Stripped(
                f"""\
// Render the path as a string.
func (se *SerializationError) PathString() string {{
{I}return aasreporting.ToGolangPath(se.Path)
}}"""
            ),
        ]
    )

    blocks.append(_generate_int64_to_jsonable())
    blocks.append(_generate_bytes_to_jsonable())

    for enum in symbol_table.enumerations:
        blocks.append(_generate_enumeration_to_jsonable(enum))

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Jsonization/{cls.name}_to_map.go"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The jsonization snippet is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue
        else:
            blocks.append(_generate_cls_to_map(cls))

    blocks.append(_generate_to_jsonable(symbol_table))

    blocks.append(Stripped("// endregion"))

    if len(errors) > 0:
        return None, errors

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate C++ code to de/wstringify enumerations and primitives."""
from aas_core_codegen.cpp.wstringification import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate C++ constants corresponding to the constants of the meta-model."""
import io
from typing import (
    Optional,
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
    description as cpp_description,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
)


# region Generation


def _generate_documentation_comment_for_constant(
    description: intermediate.DescriptionOfConstant,
    context: cpp_description.Context,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given constant."""
    # fmt: off
    comment, errors = (
        cpp_description
        .generate_comment_for_summary_remarks(
            description=description, context=context
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None
    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive_definition(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=cpp_description.Context(
                namespace=cpp_common.CONSTANTS_NAMESPACE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    constant_name = cpp_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        writer.write(f"extern const bool {constant_name};")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        writer.write(f"extern const int64_t {constant_name};")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        writer.write(f"extern const double {constant_name};")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        writer.write(f"extern const std::wstring {constant_name};")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        writer.write(f"extern const std::vector<std::uint8_t> {constant_name};")

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive_implementation(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the implementation of a constant primitive."""
    constant_name = cpp_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        assert isinstance(constant.value, bool)
        literal = cpp_common.boolean_literal(constant.value)

        return Stripped(f"const bool {constant_name} = {literal};"), None

    elif constant.a_type is intermediate.PrimitiveType.INT:
        assert isinstance(constant.value, int)

        if constant.value > 2**63 - 1 or constant.value < -(2**63):
            return None, Error(
                constant.parsed.node,
                f"The value of the constant {constant.name!r} overflows "
                f"64-bit signed integer",
            )
        return Stripped(f"const int64_t {constant_name} = {str(constant.value)};"), None

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        assert isinstance(constant.value, float)

        # NOTE (mristin, 2023-07-05):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = cpp_common.float_literal(constant.value)

        return Stripped(f"const double {constant_name} = {literal};"), None

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = cpp_common.wstring_literal(constant.value)

        return (
            Stripped(
                f"""\
const std::wstring {constant_name} = (
{I}{indent_but_first_line(literal, I)}
);"""
            ),
            None,
        )

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)

        literal, _ = cpp_common.bytes_literal(value=constant.value)

        return (
            Stripped(
                f"""\
const std::vector<std::uint_8> {constant_name} = (
{I}{indent_but_first_line(literal, I)}
);"""
            ),
            None,
        )

    else:
        assert_never(constant.a_type)


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives_definition(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    errors = []  # type: List[Error]

    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=cpp_description.Context(
                namespace=cpp_common.CONSTANTS_NAMESPACE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            errors.append(
                Error(
                    constant.parsed.node,
                    f"Failed to generate the documentation comment for {constant.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None
            writer.write(comment)
            writer.write("\n")

    set_type: Optional[str]

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        set_type = "std::unordered_set<bool>"

    elif constant.a_type is intermediate.PrimitiveType.INT:
        set_type = "std::unordered_set<int64_t>"

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        set_type = "std::unordered_set<double>"

    elif constant.a_type is intermediate.PrimitiveType.STR:
        set_type = "std::unordered_set<std::wstring>"

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        set_type = "std::unordered_set<std::vector<std::uint8_t>, HashBytes>"

    else:
        assert_never(constant.a_type)

    constant_name = cpp_naming.constant_name(constant.name)

    writer.write(
        f"""\
extern const {set_type} {constant_name};"""
    )

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives_implementation(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the implementation of a constant set of primitives."""
    literal_codes = []  # type: List[str]

    set_type: str

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        set_type = "std::unordered_set<bool>"

        for literal in constant.literals:
            assert isinstance(literal.value, bool)
            literal_codes.append(cpp_common.boolean_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.INT:
        set_type = "std::unordered_set<int64_t>"

        for literal in constant.literals:
            assert isinstance(literal.value, int)

            literal_codes.append(str(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        set_type = "std::unordered_set<double>"

        for literal in constant.literals:
            assert isinstance(literal.value, float)

            literal_codes.append(cpp_common.float_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.STR:
        set_type = "std::unordered_set<std::wstring>"

        for literal in constant.literals:
            assert isinstance(literal.value, str)
            literal_codes.append(cpp_common.wstring_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        set_type = Stripped(
            f"""\
std::unordered_set<
{I}std::vector<std::uint8_t>,
{I}HashBytes
>"""
        )

        for literal in constant.literals:
            assert isinstance(literal.value, bytearray)
            literal_code, _ = cpp_common.bytes_literal(literal.value)
            literal_codes.append(literal_code)

    else:
        assert_never(constant.a_type)

    literals_joined = ",\n".join(literal_codes)

    constant_name = cpp_naming.constant_name(constant.name)

    return (
        Stripped(
            f"""\
const {set_type} {constant_name} = {{
{I}{indent_but_first_line(literals_joined, I)}
}};"""
        ),
        None,
    )


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals_definition(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=cpp_description.Context(
                namespace=cpp_common.CONSTANTS_NAMESPACE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    enum_name = cpp_naming.enum_name(constant.enumeration.name)

    constant_name = cpp_naming.constant_name(constant.name)

    writer.write(
        f"extern const std::unordered_set<types::{enum_name}> {constant_name};"
    )

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals_implementation(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""

    enum_name = cpp_naming.enum_name(constant.enumeration.name)

    literal_codes = []  # type: List[str]
    for literal in constant.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)

        literal_codes.append(f"types::{enum_name}::{literal_name}")

    constant_name = cpp_naming.constant_name(constant.name)

    literals_joined = ",\n".join(literal_codes)

    return (
        Stripped(
            f"""\
const std::unordered_set<types::{enum_name}> {constant_name} = {{
{I}{indent_but_first_line(literals_joined, I)}
}};"""
        ),
        None,
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate C++ header code of the constants based on the symbol table."""
    errors = []  # type: List[Error]

    namespace = Stripped(f"{library_namespace}::{cpp_common.CONSTANTS_NAMESPACE}")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <cstdint>
#include <unordered_set>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup constants Pre-defined constants of the meta-model
 * @{{
 */
namespace {cpp_common.CONSTANTS_NAMESPACE} {{"""
        ),
        Stripped(
            f"""\
/**
 * Hash a blob of bytes based on the Java's String hash.
 */
struct HashBytes {{
{I}std::size_t operator()(const std::vector<std::uint8_t>& bytes) const;
}};"""
        ),
    ]

    for constant in symbol_table.constants:
        block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            block, error = _generate_constant_primitive_definition(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            block, error = _generate_constant_set_of_primitives_definition(
                constant=constant
            )
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            block, error = _generate_constant_set_of_enumeration_literals_definition(
                constant=constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert block is not None
        blocks.append(block)

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            Stripped(
                f"""\
}}  // namespace {cpp_common.COMMON_NAMESPACE}
/**@}}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate C++ implementation code of the constants based on the symbol table."""
    errors = []  # type: List[Error]

    namespace = Stripped(f"{library_namespace}::constants")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f'''\
#include "{include_prefix_path}/constants.hpp"'''
        ),
        cpp_common.WARNING,
        cpp_common.generate_namespace_opening(namespace),
        Stripped(
            f"""\
std::size_t HashBytes::operator()(
{I}const std::vector<std::uint8_t>& bytes
) const {{
{I}std::size_t result = 0;
{I}const std::size_t prime = 31;
{I}const std::size_t size = bytes.size();
{I}for (std::size_t i = 0; i < size; ++i) {{
{II}result = bytes[i] + (result * prime);
{I}}}
{I}return result;
}}"""
        ),
    ]

    for constant in symbol_table.constants:
        block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            block, error = _generate_constant_primitive_implementation(
                constant=constant
            )
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            block, error = _generate_constant_set_of_primitives_implementation(
                constant=constant
            )
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            (
                block,
                error,
            ) = _generate_constant_set_of_enumeration_literals_implementation(
                constant=constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert block is not None
        blocks.append(block)

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate the unit test code for Python."""

from aas_core_codegen.python.tests import (
    _generate_common,
    _generate_common_jsonization,
    _generate_common_xmlization,
    _generate_test_descend_and_pass_through_visitor,
    _generate_test_descend_once,
    _generate_test_jsonization_of_classes_with_descendants,
    _generate_test_jsonization_of_concrete_classes,
    _generate_test_jsonization_of_enums,
    _generate_test_for_over_x_or_empty,
    _generate_test_for_x_or_default,
    _generate_test_xmlization_of_classes_with_descendants,
    _generate_test_xmlization_of_concrete_classes,
)

generate_common = _generate_common.generate
generate_common_jsonization = _generate_common_jsonization.generate
generate_common_xmlization = _generate_common_xmlization.generate
generate_test_descend_and_pass_through_visitor = (
    _generate_test_descend_and_pass_through_visitor.generate
)
generate_test_descend_once = _generate_test_descend_once.generate
generate_test_jsonization_of_classes_with_descendants = (
    _generate_test_jsonization_of_classes_with_descendants.generate
)
generate_test_jsonization_of_concrete_classes = (
    _generate_test_jsonization_of_concrete_classes.generate
)
generate_test_jsonization_of_enums = _generate_test_jsonization_of_enums.generate
generate_test_for_over_x_or_empty = _generate_test_for_over_x_or_empty.generate
generate_test_for_x_or_default = _generate_test_for_x_or_default.generate
generate_test_xmlization_of_classes_with_descendants = (
    _generate_test_xmlization_of_classes_with_descendants.generate
)
generate_test_xmlization_of_concrete_classes = (
    _generate_test_xmlization_of_concrete_classes.generate
)



"""
Generate Go identifiers based on the identifiers from the meta-model.

The methods all generate public names (capitalized), unless their prefix indicates
otherwise.
"""
from typing import List, Iterator, Optional

from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier


@require(lambda identifier_part: len(identifier_part) > 0)
@require(lambda identifier_part: "_" not in identifier_part)
def _capitalize_or_leave_abbreviation(identifier_part: str) -> str:
    """
    Capitalize the first letter of ``text`` if it does not start with a capital letter.

    If the text already starts with a capital letter or a number, it is returned as-is.
    We thus assume that ``text`` is an abbreviation.

    >>> _capitalize_or_leave_abbreviation("something")
    'Something'

    >>> _capitalize_or_leave_abbreviation("SoMeThing")
    'SoMeThing'

    >>> _capitalize_or_leave_abbreviation("1something")
    '1something'

    >>> _capitalize_or_leave_abbreviation("URL")
    'URL'
    """
    if identifier_part[0].isupper():
        return identifier_part

    return identifier_part.capitalize()


def capital_camel_case(identifier: Identifier) -> Identifier:
    """
    Generate a capital Go camel-case name for something.

    This is usually used for public methods, members *etc.*

    >>> capital_camel_case(Identifier("something"))
    'Something'

    >>> capital_camel_case(Identifier("URL_to_something"))
    'URLToSomething'

    >>> capital_camel_case(Identifier("something_to_URL"))
    'SomethingToURL'
    """
    parts = identifier.split("_")

    return Identifier(
        "".join(
            _capitalize_or_leave_abbreviation(part) for part in parts if len(part) > 0
        )
    )


def _lower_camel_case(identifier: Identifier) -> Identifier:
    """
    Generate a lower-case Go camel-case name for something.

    This is usually used for private methods, members *etc.*

    >>> _lower_camel_case(Identifier("something"))
    'something'

    >>> _lower_camel_case(Identifier("Something"))
    'something'

    >>> _lower_camel_case(Identifier("URL_to_something"))
    'urlToSomething'

    >>> _lower_camel_case(Identifier("Something_to_URL"))
    'somethingToURL'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    parts_it = iter(parts)
    part = next(parts_it, None)
    assert part is not None, "Expected a non-empty identifier"

    cased.append(part.lower())

    part = next(parts_it, None)
    while part is not None:
        cased.append(_capitalize_or_leave_abbreviation(part))
        part = next(parts_it, None)

    return Identifier("".join(cased))


def interface_name(identifier: Identifier) -> Identifier:
    """Generate a Go public interface name based on its meta-model ``identifier``."""
    return Identifier(f"I{capital_camel_case(identifier)}")


def enum_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a public enum based on its meta-model ``identifier``.

    >>> enum_name(Identifier("something"))
    'Something'

    >>> enum_name(Identifier("URL_to_something"))
    'URLToSomething'
    """
    return capital_camel_case(identifier)


def enum_literal_name(
    enumeration_name: Identifier, literal_name: Identifier
) -> Identifier:
    """
    Generate a Go name for a public enum literal by prefixing it with the enumeration.

    >>> enum_literal_name(Identifier("URL"), Identifier("ID_short"))
    'URLIDShort'
    """
    return Identifier(
        capital_camel_case(enumeration_name) + capital_camel_case(literal_name)
    )


def private_struct_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a private struct based on its meta-model ``identifier``.

    >>> private_struct_name(Identifier("something"))
    'something'

    >>> private_struct_name(Identifier("URL_to_something"))
    'urlToSomething'

    >>> private_struct_name(Identifier("something_to_URL"))
    'somethingToURL'
    """
    return _lower_camel_case(identifier)


def struct_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a public struct based on its meta-model ``identifier``.

    >>> struct_name(Identifier("something"))
    'Something'

    >>> struct_name(Identifier("URL_to_something"))
    'URLToSomething'
    """
    return capital_camel_case(identifier)


def getter_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a public property getter based on its meta-model ``identifier``.

    >>> getter_name(Identifier("something"))
    'Something'

    >>> getter_name(Identifier("something_to_URL"))
    'SomethingToURL'
    """
    return capital_camel_case(identifier)


def setter_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a public property setter based on its meta-model ``identifier``.

    >>> setter_name(Identifier("something"))
    'SetSomething'

    >>> setter_name(Identifier("something_to_URL"))
    'SetSomethingToURL'
    """
    return capital_camel_case(Identifier("set_" + identifier))


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a public property based on its meta-model ``identifier``.

    >>> property_name(Identifier("something"))
    'Something'

    >>> property_name(Identifier("something_to_URL"))
    'SomethingToURL'
    """
    return capital_camel_case(identifier)


def private_property_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a private property based on the ``identifier``.

    If the property conflicts with the Golang keyword ``type``, it is
    translated as ``typE``. This is ugly, but we couldn't find a better
    convention that hurts less.

    >>> private_property_name(Identifier("something"))
    'something'

    >>> private_property_name(Identifier("something_to_URL"))
    'somethingToURL'

    >>> private_property_name(Identifier("type"))
    'typE'
    """
    if identifier == "type":
        return Identifier("typE")

    return _lower_camel_case(identifier)


def private_method_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a private method based on the ``identifier``.

    >>> private_method_name(Identifier("something"))
    'something'

    >>> private_method_name(Identifier("something_to_URL"))
    'somethingToURL'
    """
    return _lower_camel_case(identifier)


def method_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a member method based on its meta-model ``identifier``.

    >>> method_name(Identifier("do_something"))
    'DoSomething'

    >>> method_name(Identifier("do_something_to_URL"))
    'DoSomethingToURL'
    """
    return capital_camel_case(identifier)


def function_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a function from its meta-model ``identifier``.

    >>> function_name(Identifier("do_something"))
    'DoSomething'

    >>> function_name(Identifier("do_something_to_URL"))
    'DoSomethingToURL'
    """
    return capital_camel_case(identifier)


def private_function_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a private function from its meta-model ``identifier``.

    >>> private_function_name(Identifier("do_something"))
    'doSomething'

    >>> private_function_name(Identifier("do_something_to_URL"))
    'doSomethingToURL'
    """
    return _lower_camel_case(identifier)


def argument_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for an argument based on its meta-model ``identifier``.

    If the argument conflicts with the Golang keyword ``type``, it is
    translated as ``aType``. This is ugly, but we couldn't find a better
    convention that hurts less.

    >>> argument_name(Identifier("something"))
    'something'

    >>> argument_name(Identifier("something_to_URL"))
    'somethingToURL'

    >>> argument_name(Identifier("type"))
    'typE'
    """
    if identifier == "type":
        return Identifier("typE")

    return _lower_camel_case(identifier)


def variable_name(identifier: Identifier) -> Identifier:
    """
    Generate a Go name for a variable based on its meta-model ``identifier``.

    If the variable conflicts with the Golang keyword ``type``, it is
    translated as ``typE``. This is ugly, but we couldn't find a better
    convention that hurts less.

    >>> variable_name(Identifier("something"))
    'something'

    >>> variable_name(Identifier("something_to_URL"))
    'somethingToURL'

    >>> variable_name(Identifier("type"))
    'typE'
    """
    if identifier == "type":
        return Identifier("typE")

    return _lower_camel_case(identifier)


def constant_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a public constant based on its meta-model ``identifier``.

    >>> constant_name(Identifier("something"))
    'Something'

    >>> constant_name(Identifier("URL_to_something"))
    'URLToSomething'
    """
    return capital_camel_case(identifier)


def private_constant_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a private constant based on its meta-model ``identifier``.

    >>> private_constant_name(Identifier("something"))
    'something'

    >>> private_constant_name(Identifier("URL_to_something"))
    'urlToSomething'

    >>> private_constant_name(Identifier("something_to_URL"))
    'somethingToURL'
    """
    return _lower_camel_case(identifier)


def _over_potential_receivers(class_name: Identifier) -> Iterator[Identifier]:
    """Iterate over potential receiver names for the class."""
    yield Identifier("".join(part[0].lower() for part in class_name.split("_")))

    for i in range(2, 3):
        yield Identifier(class_name[0].lower() * i)


def receiver_name(
    cls: intermediate.ConcreteClass, prefix: Optional[str] = None
) -> Identifier:
    """
    Pick the name for the receiver variable in a method.

    Do not shadow a pre-defined set of variable names.

    If ``prefix`` is specified, append it to the receiver name.
    """
    total_arg_name_set = set(
        argument_name(arg.name) for method in cls.methods for arg in method.arguments
    )

    # NOTE (mristin, 2023-03-31):
    # The argument ``value`` is used in setters.
    total_arg_name_set.add(Identifier("value"))

    # NOTE (mristin, 2023-03-31):
    # The argument ``action`` is used in Descend and DescendOnce.
    total_arg_name_set.add(Identifier("action"))

    # NOTE (mristin, 2023-05-12):
    # The variable ``abort`` is used in Descend and DescendOnce.
    total_arg_name_set.add(Identifier("abort"))

    # NOTE (mristin, 2023-03-31):
    # The name ``enhancement`` is used as a property of an enhanced instance.
    # We skip it as a receiver to avoid confusion.
    total_arg_name_set.add(Identifier("enhancement"))

    # NOTE (mristin, 2023-03-31):
    # The name ``instance`` is used as a property of an enhanced instance.
    # We skip it as a receiver to avoid confusion.
    total_arg_name_set.add(Identifier("instance"))

    receiver = None  # type: Optional[Identifier]

    cls_name = cls.name if prefix is None else Identifier(prefix + cls.name)

    for potential_receiver in _over_potential_receivers(cls_name):
        if potential_receiver not in total_arg_name_set:
            receiver = potential_receiver
            break

    if receiver is None:
        raise AssertionError(
            f"(mristin, 2023-03-31): "
            f"No receiver name could be found for the class name {cls_name!r}. "
            f"Please contact the developers to extend the logic to add more potential "
            f"receiver names."
        )

    return receiver



"""Generate Golang code to handle AAS models based on the meta-model."""
from typing import TextIO

from aas_core_codegen import run, intermediate, specific_implementations
from aas_core_codegen.common import Stripped
from aas_core_codegen.golang import (
    aas_common as golang_aas_common,
    constants as golang_constants,
    enhancing as golang_enhancing,
    jsonization as golang_jsonization,
    reporting as golang_reporting,
    stringification as golang_stringification,
    structure as golang_structure,
    verification as golang_verification,
    xmlization as golang_xmlization,
)


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    verified_ir_table, errors = golang_structure.verify(
        symbol_table=context.symbol_table
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of Golang code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    # region Repo URL

    repo_url_key = specific_implementations.ImplementationKey("repo_url.txt")
    repo_url_text = context.spec_impls.get(repo_url_key, None)
    if repo_url_text is None:
        stderr.write(f"The repo URL snippet is missing: {repo_url_key}\n")
        return 1

    repo_url = Stripped(repo_url_text.strip())

    # endregion

    # region Common
    pth = context.output_dir / "common/common.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(golang_aas_common.generate(), encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the common Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Constants

    code, errors = golang_constants.generate(
        symbol_table=context.symbol_table, repo_url=repo_url
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the constants in the Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "constants/constants.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the constants in the Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Enhancing

    code, errors = golang_enhancing.generate(
        symbol_table=context.symbol_table,
        spec_impls=context.spec_impls,
        repo_url=repo_url,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the enhancing Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "enhancing/enhancing.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the enhancing Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Jsonization

    code, errors = golang_jsonization.generate(
        symbol_table=context.symbol_table,
        spec_impls=context.spec_impls,
        repo_url=repo_url,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the jsonization Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "jsonization/jsonization.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the jsonization Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Reporting

    code = golang_reporting.generate()

    pth = context.output_dir / "reporting/reporting.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the reporting Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Stringification

    code, errors = golang_stringification.generate(
        symbol_table=context.symbol_table, repo_url=repo_url
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the stringification Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "stringification/stringification.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the stringification Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Structure

    code, errors = golang_structure.generate(
        symbol_table=verified_ir_table,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the structures in the Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "types/types.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the Golang structures to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Verification

    verify_errors = golang_verification.verify(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if verify_errors is not None:
        run.write_error_report(
            message="Failed to verify for the generation of Golang verification code",
            errors=verify_errors,
            stderr=stderr,
        )
        return 1

    code, errors = golang_verification.generate(
        symbol_table=verified_ir_table, spec_impls=context.spec_impls, repo_url=repo_url
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the verification Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "verification/verification.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the verification Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Xmlization

    code, errors = golang_xmlization.generate(
        symbol_table=context.symbol_table,
        spec_impls=context.spec_impls,
        repo_url=repo_url,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the xmlization Golang code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "xmlization/xmlization.go"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the xmlization Golang code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Generate RDF and SHACL identifiers based on the identifiers from the meta-model."""
from typing import List

from icontract import require

from aas_core_codegen import naming
from aas_core_codegen.common import Identifier, Stripped

_LOWERCASE_WORDS_IN_LABEL = {"to", "in"}


def _capitalized_label(identifier: Identifier) -> Stripped:
    """
    Generate the label starting with a capital letter.

    >>> _capitalized_label(Identifier("Something"))
    'Something'

    >>> _capitalized_label(Identifier("Something_good_to_URL"))
    'Something Good to URL'

    >>> _capitalized_label(Identifier("URL_to_something"))
    'URL to Something'
    """
    parts = identifier.split("_")

    cased = []  # type: List[str]
    for part in parts:
        if part in _LOWERCASE_WORDS_IN_LABEL:
            cased.append(part.lower())
        else:
            if part == part.lower():
                cased.append(part.capitalize())
            else:
                cased.append(part)

    return Stripped(" ".join(cased))


# fmt: off
@require(
    lambda identifier: identifier[0].isupper(),
    "The class name must start with a capital letter"
)
# fmt: on
def class_name(identifier: Identifier) -> Identifier:
    """
    Generate the class name from the intermediate class ``identifier``.

    >>> class_name(Identifier("Something"))
    'Something'

    >>> class_name(Identifier("Something_to_URL"))
    'SomethingToUrl'

    >>> class_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


# fmt: off
@require(
    lambda identifier: identifier[0].isupper(),
    "The class name must start with a capital letter"
)
# fmt: on
def class_label(identifier: Identifier) -> Stripped:
    """
    Generate the class label from the intermediate class ``identifier``.

    >>> class_label(Identifier("Something"))
    'Something'

    >>> class_label(Identifier("Something_good_to_URL"))
    'Something Good to URL'

    >>> class_label(Identifier("URL_to_something"))
    'URL to Something'
    """
    return _capitalized_label(identifier)


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate a property name based on its meta-model ``identifier``.

    >>> property_name(Identifier("something"))
    'something'

    >>> property_name(Identifier("something_to_URL"))
    'somethingToUrl'

    >>> property_name(Identifier("URL_to_something"))
    'urlToSomething'
    """
    return naming.lower_camel_case(identifier)


def property_label(identifier: Identifier) -> Stripped:
    """
    Generate the property label from the intermediate class ``identifier``.

    >>> property_label(Identifier("something"))
    'something'

    >>> property_label(Identifier("something_good_to_URL"))
    'something good to URL'

    >>> property_label(Identifier("URL_to_something"))
    'URL to something'
    """
    return Stripped(" ".join(identifier.split("_")))


def enumeration_literal(identifier: Identifier) -> Stripped:
    """
    Generate the enumeration literal for its intermediate ``identifier``.

    >>> enumeration_literal(Identifier('something'))
    'Something'

    >>> enumeration_literal(Identifier("something_to_URL"))
    'SomethingToUrl'

    >>> enumeration_literal(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


def enumeration_literal_label(identifier: Identifier) -> Stripped:
    """
    Generate the label for an enumeration literal with intermediate ``identifier``.

    >>> enumeration_literal_label(Identifier("something"))
    'Something'

    >>> enumeration_literal_label(Identifier("something_good_to_URL"))
    'Something Good to URL'

    >>> enumeration_literal_label(Identifier("URL_to_something_good"))
    'URL to Something Good'
    """
    return _capitalized_label(identifier)
